with Ada.Text_IO;

package body SoE is

   --  each process instance gets an own Id
   Assigned_Id : Natural := 0;

   --  function specification
   --+-------------------------------------
   function Get_Id return Positive;
   --  function body
   function Get_Id return Positive is
   begin
      Assigned_Id := Assigned_Id + 1;
      return Assigned_Id;
   end Get_Id;

   --  function specification
   --+-------------------------------------
   function New_Sieve return Sieve_Ref;
   --  function body
   function New_Sieve return Sieve_Ref is
   begin
      return new Sieve_T (Get_Id);
   end New_Sieve;

   --  this is the procedure that will be called upon the finalization
   --+ of any scope that contains objects of type Last_Wishes
   --+-------------------------------------
   procedure Finalize (Last_Wishes : in out Last_Wishes_T) is
      use Ada.Text_IO;
   begin
      Put_Line ("Process instance " &
         Natural'Image (Last_Wishes.Id) &
         " has completed and can now terminate.");
   end Finalize;

   --+-------------------------------------
   task body Odd is
      --  process Odd has a child process S (instance of Sieve_T)
      --+ with which it will rendezvous for passing down the numbers
      --+ to search for primes in
      S           : Sieve_Ref;
      Num         : Positive  := 3;
      Limit       : Positive;
      Last_Wishes : Last_Wishes_T;
   begin
      --  we make sure that process Odd gets the first Id
      Last_Wishes.Id := Get_Id;
      S := New_Sieve;
      accept Set_Limit (User_Limit : Positive) do
         Limit := User_Limit;
      end Set_Limit;
      while Num < Limit loop
         --  process Odd passes all odd numbers found
         --+ in the range 3 .. Limit down to S
         --+ by means of rendezvous
         --+ (the 1st number passed down the line
         --+  is known to be a prime)
         S.Relay (Num);
         delay 0.2;
         Num := Num + 2;
      end loop;
      --  process Odd completes but cannot terminate until
      --+ its child and dependent process My_Next_Sieve has
      --+ also terminated
   end Odd;

   --+-------------------------------------
   task body Sieve_T is
      use Ada.Text_IO;
      Next_Sieve  : Sieve_Ref;
      Prime,
      Num         : Positive;
      --  any process instance of Sieve_T will possess
      --+ a finalizable object, so that its termination
      --+ will be trapped by the finalization of the object
      Last_Wishes : Last_Wishes_T;
   begin
      --  in this way, each Sieve instance will have
      --+ a finalizable object containing the process Id
      --+ of that particular instance
      Last_Wishes.Id := Id;
      --  process Sieve accepts the 1st rendezvous with its
      --+ master, knowing that the 1st incoming value will
      --+ be a prime
      accept Relay (Int : Positive) do
         Prime := Int;
      end Relay;
      Put_Line ("Sieve instance " & Natural'Image (Id) &
         " found prime number" & Natural'Image (Prime));
      loop
         --  Sieve continues to accept rendezvous with
         --+ master, to receive all other odd numbers
         --+ in the range
         --+
         --+ as we cannot be sure that there will always
         --+ be callers on this entry (the corresponding master
         --+ or parent may have completed in the meanwhile)
         --+ we should protect this process from infinite wait
         --+ and allow it to terminate in case no partner
         --+ was alive anymore
         select
            accept Relay (Int : Positive) do
               Num := Int;
            end Relay;
         or
            --  this alternative will allow the process
            --+ instance to abandon the select statement
            --+ in case no partner there were no partner
            --+ left to synchronise with
            terminate;
         end select;
         exit when Num rem Prime /= 0;
      end loop;
      --  any number which is not divisible by Prime
      --+ is itself a new prime, and we must pass it on
      --+ by rendezvous to a new child process, instance
      --+ of Sieve, so that it can continue the search
      Next_Sieve := New_Sieve;
      Next_Sieve.Relay (Num);
      loop
         --  at this point Sieve can accept from its master
         --+ all the remaining odd numbers in the range and
         --+ pass down to its child Sieve those that are
         --+ candidate primes because they are not divisible
         --+ by Prime
         --+
         --+ we must do the same as before of course
         select
            accept Relay (Int : Positive) do
               Num := Int;
            end Relay;
         or
            --  again, we need to place a terminate alternative
            --+ to allow the accepter to leave the select
            terminate;
         end select;
         if Num rem Prime /= 0 then
            Next_Sieve.Relay (Num);
         end if;
      end loop;
      --  process Sieve will complete when all select constructs
      --+ will be terminated for the lack of partners
   end Sieve_T;

end SoE;
